from typing import BinaryIO
from struct import unpack

# https://github.com/OpenL2D/moc3ingbird/blob/master/src/moc3.hexpat
class moc3:    
    Parameters : list
    Parts: list
    def __init__(self, file : BinaryIO) -> None:        
        # Header: 64 bytes
        file.seek(0)
        assert file.read(4) == b'MOC3'
        version = unpack('<c',file.read(1))[0]
        isBigEndian = unpack('<b',file.read(1))[0]
        assert not isBigEndian
                
        # TODO: Other fields
        file.seek(0x40)
        pCountInfo = unpack('<I',file.read(4))[0]
        
        file.seek(pCountInfo)
        numParts = unpack('<I',file.read(4))[0]
        file.seek(0x10, 1)
        numParameters = unpack('<I',file.read(4))[0]

        file.seek(0x4C)
        pParts = unpack('<I',file.read(4))[0]

        file.seek(0x108)
        pParameters = unpack('<I',file.read(4))[0]
        
        def read_strings(offset, count):
            for i in range(0,count):
                file.seek(offset + i * 0x40)   
                buffer = bytearray()  
                while b := file.read(1)[0]:
                    buffer.append(b)
                yield buffer.decode(encoding='utf-8')
        
        self.Parts = list(read_strings(pParts,numParts))
        self.Parameters = list(read_strings(pParameters,numParameters))



from sssekai.unity.AssetBundle import load_assetbundle
from sssekai.abcache import AbCache, AbCacheConfig,SekaiAssetBundleThreadpoolDownloader, logger, DEFAULT_CACHE_DIR
def main_mvdata(args):
    from UnityPy.enums import ClassIDType
    from pprint import pprint
    cache_dir = args.cache_dir or DEFAULT_CACHE_DIR
    config = AbCacheConfig(None, cache_dir)
    cache = AbCache(config)
    mvdata_keys = filter(lambda x: x.startswith('live_pv/mv_data/'),cache.list_entry_keys())
    mvdata_items = list()
    mvdata_lut = dict()
    for key in mvdata_keys:
        entry = cache.query_entry(key)
        if entry.get_file_exists(config):
            path = entry.get_file_path(config)
            with open(path,'rb') as f:
                env = load_assetbundle(f)
                for obj in env.objects:
                    if obj.type == ClassIDType.MonoBehaviour:
                        data = obj.read()
                        if data.name == 'data':               
                            index = len(mvdata_items)
                            mvdata_items.append(data.read_typetree())
                            mvdata_lut[str(data.name)] = index
                            mvdata_lut[str(data.id)] = index
                            break
    mvdata = mvdata_items[mvdata_lut[args.query]]
    pprint(mvdata)
    


import os
from sssekai.unity.AssetBundle import load_assetbundle
from logging import getLogger
logger = getLogger(__name__)

def main_spineextract(args):
    outdir = args.outdir
    with open(args.infile, "rb") as f:
        env = load_assetbundle(f)
        objects = [pobj.read() for pobj in env.objects]
        objects = {obj.name: obj for obj in objects if hasattr(obj,'name')}
        spines = set()
        for name in objects:
            if name.endswith(".atlas") or name.endswith(".skel"):
                spines.add(".".join(name.split(".")[:-1]))
        for spine in spines:
            logger.info('Extracting %s' % spine)
            atlas = objects.get(spine + ".atlas", None)
            skel = objects.get(spine + ".skel", None)
            texture = objects.get(spine, None)
            os.makedirs(os.path.join(outdir, spine), exist_ok=True)
            if atlas:
                with open(os.path.join(outdir, spine, spine + ".atlas.txt"), "wb") as f:
                    f.write(atlas.script)
            else:
                logger.warning("No atlas found for %s" % spine)

            if skel:
                with open(os.path.join(outdir, spine, spine + ".skel.bytes"), "wb") as f:
                    f.write(skel.script)
            else:
                logger.warning("No skel found for %s" % spine)

            if texture:            
                with open(os.path.join(outdir, spine, spine + ".png"), "wb") as f:
                    texture.image.save(f)
            else:
                logger.warning("No texture found for %s" % spine)



import os
def main_abdecrypt(args):
    from sssekai.crypto.AssetBundle import decrypt
    tree = os.walk(args.indir)
    for root, dirs, files in tree:
        for fname in files:
            file = os.path.join(root,fname)
            if (os.path.isfile(file)):
                decrypt(open(file,'rb'), open(os.path.join(args.outdir, fname),'wb'))



import os,sys
def main_mitm(args):
    from mitmproxy.tools.main import mitmdump
    current_dir = os.path.dirname(os.path.realpath(__file__))    
    args = [*sys.argv[1:], '-s', '"%s"' % os.path.join(os.path.dirname(os.path.dirname(current_dir)),'mitmproxy_sekai_api.py')]
    print('running mitmdump with args:', *args)
    mitmdump(args=args)



from typing import BinaryIO
from sssekai.crypto.AssetBundle import has_magic, decrypt

from . import UnityPy, SEKAI_UNITY_VERSION
from UnityPy import Environment

def load_assetbundle(file : BinaryIO) -> Environment:
    UnityPy.config.FALLBACK_VERSION_WARNED = True
    UnityPy.config.FALLBACK_UNITY_VERSION = SEKAI_UNITY_VERSION 
    stream = file
    if has_magic(file):
        stream = decrypt(file)
    return UnityPy.load(stream)



from sssekai.unity.AssetBundle import load_assetbundle

from os import path,remove,makedirs
def main_usmdemux(args):
    from UnityPy.enums import ClassIDType
    from wannacri.usm import Usm

    with open(args.infile,'rb') as f:
        env = load_assetbundle(f)    
        datas = dict()
        for obj in env.objects:
            if obj.type in {ClassIDType.MonoBehaviour, ClassIDType.TextAsset}:
                data = obj.read()
                datas[data.name] = data    
        movieInfo = datas.get('MovieBundleBuildData',None)
        assert movieInfo, "Invalid AssetBundle. No MovieBundleBuildData found!"
        movieInfo = movieInfo.read_typetree()
        usm_name = movieInfo['movieBundleDatas'][0]['usmFileName'][:-len('.bytes')]
        print('USM: %s' % usm_name)
        usm_folder = path.join(args.outdir,usm_name)
        makedirs(usm_folder,exist_ok=True)
        usm_temp = path.join(usm_folder,usm_name + '.tmp')
        with open(usm_temp,'wb') as usmstream:
            for data in movieInfo['movieBundleDatas']:
                usm = data['usmFileName'][:-len('.bytes')]
                usm = datas[usm]
                usmstream.write(usm.script)
        usm = Usm.open(usm_temp,encoding='shift-jis')
        usm.demux(path.join(args.outdir,usm_name),usm_name)
        remove(usm_temp)
        print('Saved to %s/' % usm_folder)


from os import path, remove, makedirs, stat
from typing import List, Mapping
from dataclasses import dataclass, asdict
from requests import Session
from logging import getLogger
from pathlib import Path
from concurrent.futures import ThreadPoolExecutor
logger = getLogger('sssekai.abcache')

from sssekai.crypto.APIManager import decrypt
from sssekai.crypto.AssetBundle import decrypt_headaer_inplace, SEKAI_AB_MAGIC
from msgpack import unpackb, dump, load
from tqdm import tqdm
DEFAULT_CACHE_DIR = '~/.sssekai/abcache'
DEFAULT_SEKAI_LATEST_VERSION = 'latest'
DEFAULT_SEKAI_PLATFORM = 'android'
DOWNLOADER_WORKER_COUNT = 8
# TODO: These seems to be in pairs?
DEFAULT_SEKAI_FALLBACK_VERSION = '3.4.1'
DEFAULT_SEKAI_FALLBACK_APP_HASH = 'a3015fe8-785f-27e1-fb8b-546a23c82c1f'

DEFAULT_FILESIZE_MATCH_RATIO = 1.5
class SekaiAssetBundleThreadpoolDownloader(ThreadPoolExecutor):
    session : Session
    progress : tqdm

    def download(self, url, fname, length):
        RETRIES = 1
        for _ in range(0,RETRIES):
            try:
                resp = self.session.get(url,stream=True)
                resp.raise_for_status()
                makedirs(path.dirname(fname),exist_ok=True)
                with open(fname, 'wb') as f:
                    magic = next(resp.iter_content(4))
                    self.progress.update(4)
                    if magic == SEKAI_AB_MAGIC:
                        header = next(resp.iter_content(128))            
                        self.progress.update(128)
                        f.write(decrypt_headaer_inplace(bytearray(header)))
                    else:
                        f.write(magic)                
                    for chunk in resp.iter_content(65536):
                        self.progress.update(len(chunk))
                        f.write(chunk)
                    return
            except Exception as e:
                logger.error('While downloading %s : %s. Retrying' % (url,e))
        if _ == RETRIES - 1:
            logger.critical('Did not download %s' % url)
    def add_link(self, url, fname, length):
        self.progress.total += length
        return self.submit(self.download, url, fname, length)
    def __init__(self, session = None) -> None:
        self.session = session or Session()
        self.progress = tqdm(bar_format="{desc}: {percentage:.1f}%|{bar}| {n_fmt}/{total_fmt} {rate_fmt} {elapsed}<{remaining}", total=0,unit='B', unit_scale=True, unit_divisor=1024,desc="Downloading")
        super().__init__(max_workers=DOWNLOADER_WORKER_COUNT)

class AbCacheConfig:
    app_version : str
    app_platform : str

    cache_dir : str # absolute path to cache directory
    downloader : SekaiAssetBundleThreadpoolDownloader

    def __init__(self, downloader : SekaiAssetBundleThreadpoolDownloader, cache_dir: str = DEFAULT_CACHE_DIR, version: str = DEFAULT_SEKAI_LATEST_VERSION, platform : str = DEFAULT_SEKAI_PLATFORM) -> None:
        self.cache_dir = path.expanduser(cache_dir)
        self.cache_dir = path.abspath(self.cache_dir)
        self.downloader = downloader
        self.app_version = version
        self.app_platform = platform

@dataclass
class AbCacheEntry(dict):
    bundleName: str
    cacheFileName: str
    cacheDirectoryName: str
    hash: str
    category: str
    crc: int
    fileSize:int
    dependencies : List[str]
    paths : List[str]
    isBuiltin : bool

    def up_to_date(self, config: AbCacheConfig, other) -> bool:
        return (
            self.hash == other.hash and 
            self.get_file_exists(config) and 
            (other.fileSize / max(1,self.get_file_size(config)) < DEFAULT_FILESIZE_MATCH_RATIO)
        )
        # HACK: we don't have proper hash checking yet. so we just check if the file exists and
        # the file size roughly matches

    def get_file_size(self, config: AbCacheConfig):
        if self.get_file_exists(config):
            return stat(self.get_file_path(config)).st_size
        else:
            return 0

    def get_file_exists(self, config: AbCacheConfig) -> bool:
        return path.exists(self.get_file_path(config))
    
    def get_file_path(self, config: AbCacheConfig) -> str:
        return path.join(config.cache_dir, self.bundleName)
@dataclass
class AbCacheIndex(dict):
    version : str = 'unknown'
    os : str = 'android'
    bundles : Mapping[str, AbCacheEntry] = None
    def get_bundle_by_abcache_path(self, config: AbCacheConfig, fpath) -> AbCacheEntry | None:
        # "C:\Users\mos9527\.sssekai\abcache\title_screen\anniversary_2nd_bg"
        fpath = Path(fpath)
        cpath = Path(config.cache_dir)
        assert fpath.is_relative_to(cpath), "File provided is not inside the cache"
        relpath = fpath.relative_to(cpath).as_posix()
        return self.bundles.get(relpath,None)    
@dataclass
class SekaiAppVersion:
    systemProfile : str
    appVersion : str
    multiPlayVersion : str
    dataVersion : str
    assetVersion : str
    appHash : str
    assetHash : str
    appVersionStatus : str
@dataclass
class SekaiSystemData:
    serverDate : int
    timezone : str
    profile : str
    maintenanceStatus : str
    appVersions : List[SekaiAppVersion]    

    def get_app_version_by_app(self, app: str) -> SekaiAppVersion | None:        
        for av in self.appVersions:
            if app in av.appVersion:
                return av
        return None
    
    def get_app_version_latest(self):
        return self.appVersions[-1]
    
    def list_app_versions(self) -> List[str]:
        return [av.appVersion for av in self.appVersions]
@dataclass
class SekaiGameVersionData:
    profile : str
    assetbundleHostHash : str
    domain : str
class AbCache(Session):    
    config : AbCacheConfig
    
    local_abcache_index : AbCacheIndex  = None

    sekai_abcache_index  : AbCacheIndex = None
    sekai_system_data : SekaiSystemData  = None
    sekai_gameversion_data : SekaiGameVersionData  = None

    @property
    def SEKAI_APP_VERSION(self): 
        if self.config.app_version == DEFAULT_SEKAI_LATEST_VERSION:
            version = self.sekai_system_data.get_app_version_latest()
        else:
            version = self.sekai_system_data.get_app_version_by_app(self.config.app_version)
        assert version, "Incorrect app version %s. Please choose from: %s" % (self.config.app_version, ', '.join(self.sekai_system_data.list_app_versions()))
        return version
    @property
    def SEKAI_ASSET_VERSION(self): return self.SEKAI_APP_VERSION.assetVersion
    @property
    def SEKAI_AB_HASH(self): return self.SEKAI_APP_VERSION.assetHash
    @property
    def SEKAI_AB_HOST_HASH(self): return self.sekai_gameversion_data.assetbundleHostHash
    @property
    def SEKAI_API_ENDPOINT(self): return 'https://production-game-api.sekai.colorfulpalette.org'
    @property
    def SEKAI_API_SYSTEM_DATA(self): return self.SEKAI_API_ENDPOINT + '/api/system'
    @property
    def SEKAI_API_GAMEVERSION_ENDPOINT(self): return 'https://game-version.sekai.colorfulpalette.org'
    @property
    def SEKAI_AB_INFO_ENDPOINT(self): return f'https://production-{self.SEKAI_AB_HOST_HASH}-assetbundle-info.sekai.colorfulpalette.org/'
    @property
    def SEKAI_AB_ENDPOINT(self): return f'https://production-{self.SEKAI_AB_HOST_HASH}-assetbundle.sekai.colorfulpalette.org/'
    @property
    def SEKAI_AB_BASE_PATH(self): return f'{self.SEKAI_ASSET_VERSION}/{self.SEKAI_AB_HASH}/{self.config.app_platform}/'
    @property
    def SEKAI_AB_INDEX_PATH(self): return f'api/version/{self.SEKAI_ASSET_VERSION}/os/{self.config.app_platform}'
    @property
    def SEKAI_ISSUE_ENDPOINT(self): return 'https://issue.sekai.colorfulpalette.org'
    @property
    def SEKAI_ISSUE_SIGNATURE_ENDPOINT(self): return self.SEKAI_ISSUE_ENDPOINT + '/api/signature'

    def update_signatures(self):
        logger.info('Updating signatures')
        resp = self.post(self.SEKAI_ISSUE_SIGNATURE_ENDPOINT)
        resp.raise_for_status()
        self.headers['Cookie'] = resp.headers['Set-Cookie'] 
        # HACK: Per RFC6265, Cookies should not be visible to subdomains since it's not set with Domain attribute (https://github.com/psf/requests/issues/2576)
        # But the other endpoints uses it nontheless. So we have to set it manually.

    def update_system_data(self):
        logger.info('Updating system data')
        resp = self.get(self.SEKAI_API_SYSTEM_DATA)
        resp.raise_for_status()
        data = decrypt(resp.content)
        data = unpackb(data)
        self.sekai_system_data = SekaiSystemData(**data)
        for i, appVersion in enumerate(self.sekai_system_data.appVersions):
            self.sekai_system_data.appVersions[i] = SekaiAppVersion(**appVersion)

    def update_gameversion_data(self):
        logger.info('Updating game version data')
        resp = self.get(self.SEKAI_API_GAMEVERSION_ENDPOINT + '/' + self.SEKAI_APP_VERSION.appVersion + '/' + self.SEKAI_APP_VERSION.appHash)
        resp.raise_for_status()
        data = decrypt(resp.content)
        data = unpackb(data)
        self.sekai_gameversion_data = SekaiGameVersionData(**data)
    
    def update_abcache_index(self) -> AbCacheIndex:
        logger.info('Updating Assetbundle index')
        resp = self.get(
            url=self.SEKAI_AB_INFO_ENDPOINT + self.SEKAI_AB_INDEX_PATH
        )
        resp.raise_for_status()
        data = decrypt(resp.content)
        data = unpackb(data)
        self.sekai_abcache_index = cache = AbCacheIndex(**data)
        for k,v in cache.bundles.items():
            cache.bundles[k] = AbCacheEntry(**v)         

    def list_entry_keys(self) -> List[str]:
        return self.local_abcache_index.bundles.keys()

    def query_entry(self, bundleName : str) -> AbCacheEntry | None:
        return self.local_abcache_index.bundles.get(bundleName, None)

    def queue_update_cache_entry(self, entry : AbCacheEntry, new_entry : AbCacheEntry = None):
        if new_entry:
            entry = new_entry
        return self.config.downloader.add_link(
            self.SEKAI_AB_ENDPOINT + self.SEKAI_AB_BASE_PATH + entry.bundleName,
            entry.get_file_path(self.config),
            entry.fileSize
        )

    def queue_update_cache_entry_full(self):        
        all_keys = sorted(list({k for k in self.sekai_abcache_index.bundles.keys()}.union({k for k in self.local_abcache_index.bundles.keys()})))
        update_count = 0
        for k in all_keys:
            if k in self.sekai_abcache_index.bundles and k in self.local_abcache_index.bundles:
                # update
                if not self.sekai_abcache_index.bundles[k].up_to_date(self.config,self.local_abcache_index.bundles[k]):
                    logger.debug('Updating bundle %s. (size on disk=%d, reported=%d)' % (k, self.local_abcache_index.bundles[k].get_file_size(self.config), self.local_abcache_index.bundles[k].fileSize))
                    self.queue_update_cache_entry(self.local_abcache_index.bundles[k], self.sekai_abcache_index.bundles[k])
                    update_count+=1
                else:                    
                    pass
            elif k in self.sekai_abcache_index.bundles: 
                # append
                logger.debug('Adding bundle %s', k)
                self.local_abcache_index.bundles[k] = self.sekai_abcache_index.bundles[k]
                self.queue_update_cache_entry(self.local_abcache_index.bundles[k])
                update_count+=1
            else: 
                # removal
                logger.debug('Removing bundle %s', k)
                if path.exists(self.local_abcache_index.bundles[k].get_file_path(self.config)):
                    remove(self.local_abcache_index.bundles[k].get_file_path(self.config))
                del self.local_abcache_index.bundles[k]
        logger.debug('Saving AssetBundle index')
        self.save()
        logger.debug('Queued %d updates' % update_count)
    
    def update_metadata(self):
        logger.info('Updating metadata')
        logger.debug('Cache directory: %s' % self.config.cache_dir)
        logger.debug('Set App version: %s (%s)' % (self.config.app_version,self.config.app_platform))
        self.update_signatures()
        self.update_system_data()
        self.update_gameversion_data()               
        self.update_abcache_index()
        # Update to the actually usable set version
        self.config.app_version = self.SEKAI_APP_VERSION.appVersion
        self.headers['X-App-Version'] = self.SEKAI_APP_VERSION.appVersion
        self.headers['X-App-Hash'] = self.SEKAI_APP_VERSION.appHash
        logger.debug('Actual App version: %s (%s), hash=%s' % (self.config.app_version,self.config.app_platform, self.SEKAI_APP_VERSION.appHash))
        logger.debug('Sekai AssetBundle version: %s' % self.SEKAI_ASSET_VERSION)
        logger.debug('Sekai AssetBundle host hash: %s' % self.SEKAI_AB_HOST_HASH)

    def __init__(self, config : AbCacheConfig) -> None:
        super().__init__()
        self.config = config
        self.headers.update({
            'Accept': 'application/octet-stream',
            'Content-Type': 'application/octet-stream',
            'Accept-Encoding': 'deflate, gzip',
            'User-Agent': 'UnityPlayer/2020.3.32f1 (UnityWebRequest/1.0, libcurl/7.80.0-DEV)',
            'X-Platform': self.config.app_platform,
            'X-Unity-Version': '2020.3.32f1',
            'X-App-Version': DEFAULT_SEKAI_FALLBACK_VERSION, # These will be updated later
            'X-App-Hash': DEFAULT_SEKAI_FALLBACK_APP_HASH
        })
        makedirs(self.config.cache_dir,exist_ok=True)
        try:
            self.load()
        except Exception as e:
            logger.warning('Failed to load cache index. Creating a new one. (%s)' % e)
            self.local_abcache_index = AbCacheIndex(bundles=dict())
            self.save()
        
    
    def save(self):
        with open(path.join(self.config.cache_dir, 'abindex'),'wb') as f:
            dump(asdict(self.local_abcache_index), f)
            logger.info("Saved %d entries to cache index" % len(self.local_abcache_index.bundles))

    def load(self):
        with open(path.join(self.config.cache_dir, 'abindex'), 'rb') as f:
            self.local_abcache_index = AbCacheIndex(**load(f, raw=False))
            for k,v in self.local_abcache_index.bundles.items():
                self.local_abcache_index.bundles[k] = AbCacheEntry(**v)
            logger.info("Loaded %d entries from cache index" % len(self.local_abcache_index.bundles))



import codecs
import sys, os
import argparse
from sssekai.entrypoint.apidecrypt import main_apidecrypt
from sssekai.entrypoint.abdecrypt import main_abdecrypt
from sssekai.entrypoint.mitm import main_mitm
from sssekai.entrypoint.mvdata import main_mvdata
from sssekai.entrypoint.usmdemux import main_usmdemux
from sssekai.entrypoint.abcache import main_abcache
from sssekai.entrypoint.live2dextract import main_live2dextract
from sssekai.entrypoint.spineextract import main_spineextract
from sssekai.unity import SEKAI_UNITY_VERSION
from sssekai.abcache import DEFAULT_CACHE_DIR, DEFAULT_SEKAI_LATEST_VERSION, DEFAULT_SEKAI_PLATFORM
def __main__():
    from tqdm.std import tqdm as tqdm_c
    class SemaphoreStdout:
        @staticmethod
        def write(__s):
            # Blocks tqdm's output until write on this stream is done
            # Solves cases where progress bars gets re-rendered when logs
            # spews out too fast
            with tqdm_c.external_write_mode(file=sys.stdout, nolock=False):
                return sys.stdout.write(__s)
    parser = argparse.ArgumentParser(description='''SSSekai Proejct SEKAI feat. Hatsune Miku (Android) Modding Tools
Installation:
    pip install git+https://github.com/mos9527/sssekai                                    
''', formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('--log-level', type=str, help='logging level (default: %(default)s)', default='INFO', choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'])
    subparsers = parser.add_subparsers(title='subcommands', description='valid subcommands', help='additional help')
    # apidecrypt
    apidecrypt_parser = subparsers.add_parser('apidecrypt', help='''API crypto dumper
This crypto applies to:
    - API request/response body dumped by packet sniffer (mitmproxy, wireshark, etc.)
    - AssetBundleInfo (can be found at /sdcard/Android/data/com.hermes.mk.asia/files/data/AssetBundleInfo,or see sssekai.abcache)''')
    apidecrypt_parser.add_argument('infile', type=str, help='input dump file')
    apidecrypt_parser.add_argument('outfile', type=str, help='output json file')
    apidecrypt_parser.set_defaults(func=main_apidecrypt)
    # abdecrypt
    abdecrypt_parser = subparsers.add_parser('abdecrypt', help='''Decrypt Sekai AssetBundle
These can be found at /sdcard/Android/data/com.hermes.mk.asia/files/data/                                             
''')
    abdecrypt_parser.add_argument('indir', type=str, help='input directory')
    abdecrypt_parser.add_argument('outdir', type=str, help='output directory')
    abdecrypt_parser.set_defaults(func=main_abdecrypt)
    # usmdemux
    usmdemux_parser = subparsers.add_parser('usmdemux', help='''Demux Sekai USM Video in a AssetBundle''')
    usmdemux_parser.add_argument('infile', type=str, help='input file')
    usmdemux_parser.add_argument('outdir', type=str, help='output directory')
    usmdemux_parser.set_defaults(func=main_usmdemux)
    # abcache
    abcache_parser = subparsers.add_parser('abcache', help='''Sekai AssetBundle local cache
Downloads/Updates *ALL* PJSK JP assets to local devices.
NOTE: The assets can take quite a lot of space (est. 42.5GB for app version 3.3.1) so be prepared
NOTE: The AssetBundles *cached* are NOT OBFUSCATED. They can be used as is by various Unity ripping tools (and sssekai by extension)
      that supports stripped Unity version (should be %s. the version is ripped).''' % SEKAI_UNITY_VERSION)
    abcache_parser.add_argument('--cache-dir', type=str, help='cache directory (default: %(default)s)',default=DEFAULT_CACHE_DIR)
    abcache_parser.add_argument('--skip-update',action='store_true',help='skip all updates and use cached assets as is.')
    abcache_parser.add_argument('--version', type=str, help='PJSK app version (default: %(default)s)', default=DEFAULT_SEKAI_LATEST_VERSION)
    abcache_parser.add_argument('--platform', type=str, help='PJSK app platform (default: %(default)s)', default=DEFAULT_SEKAI_PLATFORM)
    abcache_parser.add_argument('--open', action='store_true',help='open cache directory. this will skip all updates.')
    abcache_parser.set_defaults(func=main_abcache)
    # live2dextract
    live2dextract_parser = subparsers.add_parser('live2dextract', help='''Extract Sekai Live2D Models in a AssetBundle''')
    live2dextract_parser.add_argument('infile', type=str, help='input file')
    live2dextract_parser.add_argument('outdir', type=str, help='output directory')
    live2dextract_parser.add_argument('--no-anim',action='store_true',help='don\'t extract animation clips')
    live2dextract_parser.set_defaults(func=main_live2dextract)
    # spineextract
    spineextract_parser = subparsers.add_parser('spineextract', help='''Extract Sekai Spine (Esoteric Spine2D) Models in a AssetBundle''')
    spineextract_parser.add_argument('infile', type=str, help='input file')
    spineextract_parser.add_argument('outdir', type=str, help='output directory')    
    spineextract_parser.set_defaults(func=main_spineextract)    
    # mvdata
    mvdata_parser = subparsers.add_parser('mvdata', help='''Query Sekai MV data from AssetBundle''')
    mvdata_parser.add_argument('--cache-dir', type=str, help='abcache cache directory (default: %(default)s)',default=DEFAULT_CACHE_DIR)
    mvdata_parser.add_argument('query', type=str, help='query string. Either MV ID or MV (full) name')
    mvdata_parser.set_defaults(func=main_mvdata)
    # mitm
    mitm_parser = subparsers.add_parser('mitm', help='Run Sekai API MITM proxy (WIP)')
    mitm_parser.set_defaults(func=main_mitm)
    # parse args
    args = parser.parse_args()
    # set logging level
    import coloredlogs
    from logging import basicConfig
    coloredlogs.install(
            level=args.log_level,
            fmt="%(asctime)s %(name)s [%(levelname).4s] %(message)s",
            isatty=True,
            stream=SemaphoreStdout
        )
    basicConfig(
        level=args.log_level, format="[%(levelname).4s] %(name)s %(message)s", stream=SemaphoreStdout
    )
    if 'func' in args:
        args.func(args)
    else:
        parser.print_help()


if __name__ == "__main__":
    __main__()
    sys.exit(0)



from sssekai.abcache import AbCache, AbCacheConfig,SekaiAssetBundleThreadpoolDownloader, logger, DEFAULT_CACHE_DIR
def main_abcache(args):
    downloader = SekaiAssetBundleThreadpoolDownloader()
    cache_dir = args.cache_dir or DEFAULT_CACHE_DIR
    config = AbCacheConfig(downloader, cache_dir, args.version, args.platform)
    if args.open:
        from os import startfile
        startfile(config.cache_dir)
        return
    cache = AbCache(config)
    downloader.session = cache
    if not args.skip_update:
        cache.update_metadata()
        cache.queue_update_cache_entry_full()
        downloader.shutdown(wait=True)
    cache.save()
    logger.info('AssetBundle cache is now ready. Visit: %s' % cache_dir)
